@@ -1,4 +1,4 @@
-"""Digital Elevation Map (DEM) downloading/stitching/upsampling
+"""Digital Elevation Map (DEM) downloading, stitching, upsampling
 
 Module contains utilities for downloading all necessary .hgt files
 for a lon/lat rectangle, stiches them into one DEM, and creates a
@@ -38,15 +38,17 @@ Example .dem.rsc (for N19W156.hgt and N19W155.hgt stitched horizontally):
         PROJECTION    LL
 """
 from __future__ import division, print_function
+
 import collections
 import logging
 import os
+
 import numpy as np
 
-from sardem import utils, loading, upsample_cy, conversions
-from sardem.download import Tile, Downloader
+from sardem import conversions, loading, upsample, utils
+from sardem.constants import DEFAULT_RES, NUM_PIXELS_SRTM1
+from sardem.download import Downloader, Tile
 
-NUM_PIXELS = 3601  # For SRTM1
 RSC_KEYS = [
     "WIDTH",
     "FILE_LENGTH",
@@ -79,7 +81,7 @@ class Stitcher:
     """
 
     def __init__(
-        self, tile_names, filenames=[], data_source="NASA", num_pixels=NUM_PIXELS
+        self, tile_names, filenames=[], data_source="NASA", num_pixels=NUM_PIXELS_SRTM1
     ):
         """List should come from Tile.srtm1_tile_names()"""
         self.tile_file_list = list(tile_names)
@@ -191,7 +193,7 @@ class Stitcher:
             else:
                 return loading.load_elevation(filename)
         else:
-            return np.zeros((NUM_PIXELS, NUM_PIXELS), dtype=self.dtype)
+            return np.zeros((NUM_PIXELS_SRTM1, NUM_PIXELS_SRTM1), dtype=self.dtype)
 
     def load_and_stitch(self):
         """Function to load combine .hgt tiles
@@ -279,37 +281,13 @@ class Stitcher:
         return rsc_dict
 
 
-def crop_stitched_dem(bounds, stitched_dem, rsc_data):
-    """Takes the output of Stitcher.load_and_stitch, crops to bounds
-
-    Args:
-        bounds (tuple[float]): (left, bot, right, top) lats and lons of
-            desired bounding box for the DEM
-        stitched_dem (numpy.array, 2D): result from files
-            through Stitcher.load_and_stitch()
-        rsc_data (dict): data from .dem.rsc file, from Stitcher.create_dem_rsc
-
-    Returns:
-        numpy.array: a cropped version of the bigger stitched_dem
-    """
-    indexes, new_starts = utils.find_bounding_idxs(
-        bounds,
-        rsc_data["X_STEP"],
-        rsc_data["Y_STEP"],
-        rsc_data["X_FIRST"],
-        rsc_data["Y_FIRST"],
-    )
-    left_idx, bot_idx, right_idx, top_idx = indexes
-    cropped_dem = stitched_dem[top_idx:bot_idx, left_idx:right_idx]
-    new_sizes = cropped_dem.shape
-    return cropped_dem, new_starts, new_sizes
+def _float_is_on_bounds(x):
+    return int(x) == x
 
 
 def main(
-    left_lon=None,
-    top_lat=None,
-    dlon=None,
-    dlat=None,
+    output_name=None,
+    bbox=None,
     geojson=None,
     wkt_file=None,
     data_source=None,
@@ -318,96 +296,97 @@ def main(
     make_isce_xml=False,
     keep_egm=False,
     shift_rsc=False,
-    output_name=None,
+    cache_dir=None,
 ):
     """Function for entry point to create a DEM with `sardem`
 
     Args:
-        left_lon (float): Left most longitude of DEM box
-        top_lat (float): Top most longitude of DEM box
-        dlon (float): Width of box in longitude degrees
-        dlat (float): Height of box in latitude degrees
-        geojson (dict): geojson object outlining DEM (alternative to lat/lon)
+        output_name (str): name of file to save final DEM (default = elevation.dem)
+        bbox (tuple[float]): (left, bot, right, top)
+            Longitude/latitude desired bounding box for the DEM
+        geojson (dict): geojson object outlining DEM (alternative to bbox)
+        wkt_file (str): path to .wkt file outlining DEM (alternative to bbox)
         data_source (str): 'NASA' or 'AWS', where to download .hgt tiles from
         xrate (int): x-rate (columns) to upsample DEM (positive int)
         yrate (int): y-rate (rows) to upsample DEM (positive int)
         make_isce_xml (bool): whether to make an isce2-compatible XML file
         keep_egm (bool): Don't convert the DEM heights from geoid heights
-            above EGM96 or EGM2008 to heights above WGS84 ellipsoid (default = False)
+            above EGM96 or EGM2008 to heights above WGS84 ellipsoid 
+            (default = False: do the conversion)
         shift_rsc (bool): Shift the .dem.rsc file down/right so that the
             X_FIRST and Y_FIRST values represent the pixel *center* (instead of
             GDAL's convention of pixel edge). Default = False.
-        output_name (str): name of file to save final DEM (default = elevation.dem)
+        cache_dir (str): directory to cache downloaded tiles
     """
-    if geojson:
-        bounds = utils.bounding_box(geojson=geojson)
-    elif wkt_file:
-        bounds = utils.get_wkt_bbox(wkt_file)
-    else:
-        bounds = utils.bounding_box(left_lon, top_lat, dlon, dlat)
-    logger.info("Bounds: %s", " ".join(str(b) for b in bounds))
-    outrows, outcols = utils.get_output_size(bounds, xrate, yrate)
+    if bbox is None:
+        if geojson:
+            bbox = utils.bounding_box(geojson=geojson)
+        elif wkt_file:
+            bbox = utils.get_wkt_bbox(wkt_file)
+
+    if bbox is None:
+        raise ValueError("Must provide either bbox or geojson or wkt_file")
+    logger.info("Bounds: %s", " ".join(str(b) for b in bbox))
+
+    # if all(_float_is_on_bounds(b) for b in bbox):
+    #     logger.info("Shifting bbox to nearest tile bounds")
+    #     bbox = utils.shift_integer_bbox(bbox)
+    #     logger.info("New edge bounds: %s", " ".join(str(b) for b in bbox))
+    # Now we're assuming that `bbox` refers to the edges of the desired bounding box
+
+    # Print a warning if they're possibly requesting too-large a box by mistake
+    outrows, outcols = utils.get_output_size(bbox, xrate, yrate)
     if outrows * outcols > WARN_LIMIT:
         logger.warning(
             "Caution: Output size is {} x {} pixels.".format(outrows, outcols)
         )
-        logger.warning("Are the bounds correct?")
-
-    # Are we using GDAL's convention (pixel edge) or the center?
-    # i.e. if `shift_rsc` is False, then we are `using_gdal_bounds`
-    using_gdal_bounds = not shift_rsc
+        logger.warning("Are the bounds correct (left, bottom, right, top)?")
 
+    # For copernicus, use GDAL to warp from the VRT
     if data_source == "COP":
         utils._gdal_installed_correctly()
         from sardem import cop_dem
 
         cop_dem.download_and_stitch(
             output_name,
-            bounds,
+            bbox,
             keep_egm=keep_egm,
             xrate=xrate,
             yrate=yrate,
         )
         if make_isce_xml:
             logger.info("Creating ISCE2 XML file")
-            utils.gdal2isce_xml(
-                output_name, keep_egm=keep_egm, using_gdal_bounds=using_gdal_bounds
-            )
+            utils.gdal2isce_xml(output_name, keep_egm=keep_egm)
         return
 
-    tile_names = list(Tile(*bounds).srtm1_tile_names())
+    # If using SRTM, download tiles manually and stitch
+    tile_names = list(Tile(*bbox).srtm1_tile_names())
 
-    d = Downloader(tile_names, data_source=data_source)
+    d = Downloader(tile_names, data_source=data_source, cache_dir=cache_dir)
     local_filenames = d.download_all()
 
     s = Stitcher(tile_names, filenames=local_filenames, data_source=data_source)
     stitched_dem = s.load_and_stitch()
 
-    # Now create corresponding rsc file
-    rsc_dict = s.create_dem_rsc()
+    # Now create corresponding rsc file for all the tiles
+    rsc_dict_tiles = s.create_dem_rsc()
 
-    # Cropping: get very close to the bounds asked for:
     logger.info("Cropping stitched DEM to boundaries")
-    stitched_dem, new_starts, new_sizes = crop_stitched_dem(
-        bounds, stitched_dem, rsc_dict
-    )
-    new_x_first, new_y_first = new_starts
-    new_rows, new_cols = new_sizes
-    # Now adjust the .dem.rsc data to reflect new top-left corner and new shape
-    rsc_dict["X_FIRST"] = new_x_first
-    rsc_dict["Y_FIRST"] = new_y_first
-    rsc_dict["FILE_LENGTH"] = new_rows
-    rsc_dict["WIDTH"] = new_cols
-    if shift_rsc:
-        rsc_dict = utils.shift_rsc_dict(rsc_dict, to_gdal=True)
+    stitched_dem = upsample.resample(stitched_dem, rsc_dict_tiles, bbox)
+    rsc_dict = rsc_dict_tiles.copy()
+    rsc_dict["X_FIRST"] = bbox[0]
+    rsc_dict["Y_FIRST"] = bbox[3]
+    rsc_dict["FILE_LENGTH"] = stitched_dem.shape[0]
+    rsc_dict["WIDTH"] = stitched_dem.shape[1]
 
     rsc_filename = output_name + ".rsc"
 
     # Upsampling:
+    dtype = np.int16
     if xrate == 1 and yrate == 1:
         logger.info("Rate = 1: No upsampling to do")
         logger.info("Writing DEM to %s", output_name)
-        stitched_dem.astype(np.uint16).tofile(output_name)
+        stitched_dem.astype(dtype).tofile(output_name)
         logger.info("Writing .dem.rsc file to %s", rsc_filename)
         with open(rsc_filename, "w") as f:
             f.write(loading.format_dem_rsc(rsc_dict))
@@ -419,8 +398,7 @@ def main(
         rsc_filename_small = "small_" + rsc_filename
 
         logger.info("Writing non-upsampled dem temporarily to %s", dem_filename_small)
-        # Note: forcing to uint16 to simplify c-program loading
-        stitched_dem.astype(np.uint16).tofile(dem_filename_small)
+        stitched_dem.astype(dtype).tofile(dem_filename_small)
         logger.info(
             "Writing non-upsampled dem.rsc temporarily to %s", rsc_filename_small
         )
@@ -430,21 +408,35 @@ def main(
         # Redo a new .rsc file for it
         logger.info("Writing new upsampled dem.rsc to %s", rsc_filename)
         with open(rsc_filename, "w") as f:
-            upsampled_rsc = utils.upsample_dem_rsc(
+            upsampled_rsc = upsample.upsample_dem_rsc(
                 xrate=xrate, yrate=yrate, rsc_dict=rsc_dict
             )
             f.write(upsampled_rsc)
 
-        # Now upsample this block
-        nrows, ncols = stitched_dem.shape
-        upsample_cy.upsample_wrap(
-            dem_filename_small.encode("utf-8"),
-            xrate,
-            yrate,
-            ncols,
-            nrows,
-            output_name.encode(),
-        )
+        # Now upsample using with GDAL or python
+        if utils._gdal_installed_correctly():
+            logger.info("Using GDAL Translate for upsampling")
+            upsample.upsample_with_gdal(
+                dem_filename_small,
+                output_name,
+                method="bilinear",  # make this an option?
+                xrate=xrate,
+                yrate=yrate,
+            )
+        else:
+            # Figure out size of row blocks to keep memory under 100 MB
+            nrows, ncols = stitched_dem.shape
+            block_rows = int(np.round(10e6 / ncols / 2, -2))  # round to 100s
+            logger.info("Upsampling by blocks of {} rows".format(block_rows))
+            upsample.upsample_by_blocks(
+                dem_filename_small,
+                output_name,
+                (nrows, ncols),
+                block_rows=block_rows,
+                dtype=dtype,
+                xrate=xrate,
+                yrate=yrate,
+            )
         # Clean up the _small versions of dem and dem.rsc
         logger.info("Cleaning up %s and %s", dem_filename_small, rsc_filename_small)
         os.remove(dem_filename_small)
@@ -452,21 +444,21 @@ def main(
 
     if make_isce_xml:
         logger.info("Creating ISCE2 XML file")
-        utils.gdal2isce_xml(
-            output_name, keep_egm=keep_egm, using_gdal_bounds=using_gdal_bounds
-        )
+        utils.gdal2isce_xml(output_name, keep_egm=keep_egm)
 
-    if keep_egm or data_source == "NASA_WATER":
+    if data_source == "NASA_WATER":
+        logger.info("Water mask requires no geoid correction.")
+        # Overwrite with smaller dtype for water mask
+        upsampled_dict = loading.load_dem_rsc(rsc_filename)
+        rows, cols = upsampled_dict["file_length"], upsampled_dict["width"]
+        mask = np.fromfile(output_name, dtype=dtype).reshape((rows, cols))
+        mask.astype(bool).tofile(output_name)
+    elif keep_egm:
         logger.info("Keeping DEM as EGM96 geoid heights")
     else:
         logger.info("Correcting DEM to heights above WGS84 ellipsoid")
-        conversions.convert_dem_to_wgs84(
-            output_name, using_gdal_bounds=using_gdal_bounds
-        )
-
-    # Overwrite with smaller dtype for water mask
-    if data_source == "NASA_WATER":
-        upsampled_dict = loading.load_dem_rsc(rsc_filename)
-        rows, cols = upsampled_dict["file_length"], upsampled_dict["width"]
-        mask = np.fromfile(output_name, dtype=np.int16).reshape((rows, cols))
-        mask.astype(bool).tofile(output_name)
+        conversions.convert_dem_to_wgs84(output_name, geoid="egm96")
+ 
+    # If the user wants the .rsc file to point to pixel center:
+    if shift_rsc:
+        utils.shift_rsc_file(rsc_filename, to_gdal=False)
\ No newline at end of file
